from django.utils.translation import gettext_lazy as _

from netbox.registry import registry
from netbox.navigation import *
from netbox.navigation.menu import MENUS


PHONE = Menu(
    label=_('Phone'),
    icon_class="mdi mdi-phone",
    groups=(
        MenuGroup(
            label=_('Phone'),
            items=(
                MenuItem(
                    link=f'plugins:sop_phone:phoneinfo_list',
                    link_text=_('Informations'),
                    permissions=[f'sop_phone.view_phoneinfo'],
                    buttons=(
                        MenuItemButton(
                            link=f'plugins:sop_phone:phoneinfo_add',
                            title='Add',
                            icon_class='mdi mdi-plus-thick',
                            permissions=[f'sop_phone.add_phoneinfo'],
                        ),
                    ),
                ),
                MenuItem(
                    link=f'plugins:sop_phone:phonedelivery_list',
                    link_text=_('Deliveries'),
                    permissions=[f'sop_phone.view_phonedelivery'],
                    buttons=(
                        MenuItemButton(
                            link=f'plugins:sop_phone:phonedelivery_add',
                            title='Add',
                            icon_class='mdi mdi-plus-thick',
                            permissions=[f'sop_phone.add_phonedelivery'],
                        ),
                    ),
                ),
                MenuItem(
                    link=f'plugins:sop_phone:phonedid_list',
                    link_text=_('DIDs'),
                    permissions=[f'sop_phone.view_phonedid'],
                    buttons=(
                        MenuItemButton(
                            link=f'plugins:sop_phone:phonedid_add',
                            title='Add',
                            icon_class='mdi mdi-plus-thick',
                            permissions=[f'sop_phone.add_phonedid'],
                        ),
                        MenuItemButton(
                            link=f'plugins:sop_phone:phonedid_import',
                            title='Import',
                            icon_class='mdi mdi-upload',
                            permissions=[f'sop_phone.add_phonedid'],
                        ),
                    ),
                ),
                MenuItem(
                    link=f'plugins:sop_phone:phonemaintainer_list',
                    link_text=_('Maintainers'),
                    permissions=[f'sop_phone.view_phonemaintainer'],
                    buttons=(
                        MenuItemButton(
                            link=f'plugins:sop_phone:phonemaintainer_add',
                            title='Add',
                            icon_class='mdi mdi-plus-thick',
                            permissions=[f'sop_phone.add_phonemaintainer'],
                        ),
                        MenuItemButton(
                            link=f'plugins:sop_phone:phonemaintainer_import',
                            title='Import',
                            icon_class='mdi mdi-upload',
                            permissions=[f'sop_phone.add_phonemaintainer'],
                        ),
                    ),
                ),
            ),
        ),
    ),
)

MENUS.append(PHONE)



from django import forms
from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _

from netbox.forms import NetBoxModelFilterSetForm, NetBoxModelBulkEditForm, NetBoxModelForm, NetBoxModelImportForm
from utilities.forms.fields import DynamicModelChoiceField, DynamicModelChoiceField, CSVModelChoiceField
from utilities.forms.rendering import FieldSet
from dcim.models import Site, Region, SiteGroup

from ..models import *


__all__ = (
    'PhoneDIDForm',
    'PhoneDIDFilterForm',
    'PhoneDIDBulkEditForm',
    'PhoneDIDBulkImportForm',
)


class PhoneDIDBulkEditForm(NetBoxModelBulkEditForm):
    model = PhoneDID

    site = DynamicModelChoiceField(
        queryset=Site.objects.all()
    )
    # this field depends on the site field
    # -> you can only enter deliveries that 
    # -> share the same site as the field site   
    delivery = DynamicModelChoiceField(
        queryset=PhoneDelivery.objects.all(),
        help_text=_('Specify how this range is delivered.'),
        query_params={
            'site_id': '$site'
        }
    )

    class Meta:
        fields = ('site', 'delivery', )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if 'add_tags' in self.fields:
            del self.fields['add_tags']
        if 'remove_tags' in self.fields:
            del self.fields['remove_tags']


class PhoneDIDFilterForm(NetBoxModelFilterSetForm):
    model = PhoneDID
    
    site_id = DynamicModelChoiceField(
        queryset=Site.objects.all(),
        required=False,
        label=_('Site')
    )
    group_id = DynamicModelChoiceField(
        queryset=SiteGroup.objects.all(),
        required=False,
        label=_('Site group')
    )
    region_id = DynamicModelChoiceField(
        queryset=Region.objects.all(),
        required=False,
        label=_('Region')
    )
    maintainer_id = forms.ModelChoiceField(
        queryset=PhoneMaintainer.objects.all(),
        required=False,
        label=_('Maintainer')
    )
    delivery_id = forms.ModelChoiceField(
        queryset=PhoneDelivery.objects.all(),
        required=False,
        label=_('Delivery')
    )
    partial_number = forms.IntegerField(
        label=_('Partial number'),
        required=False,
        help_text=_('E164 format')
    )
    start = forms.IntegerField(
        label=_('Start number'),
        required=False,
        help_text=_('E164 format'),
    )
    end = forms.IntegerField(
        label=_('End number'),
        required=False,
        help_text=_('E164 format')
    )

    fieldsets = (
        FieldSet(
            'region_id', 'group_id', 'site_id',
            name=_('Location')
        ),
        FieldSet(
            'maintainer_id', 'delivery_id',
            name=_('Information')
        ),
        FieldSet(
            'partial_number', 'start', 'end',
            name=_('Attributes')
        )
    )

    def clean(self):
        super().clean()


class PhoneDIDForm(NetBoxModelForm):

    site = DynamicModelChoiceField(
        label=_('Site'),
        queryset=Site.objects.all(),
        required=True,
    )
    start = forms.IntegerField(
        label=_('Start number'),
        required=True,
        help_text=_('E164 format / NUMBER ONLY'),
    )
    end = forms.IntegerField(
        label=_('End number'),
        required=False,
        help_text=_('E164 format - can be left blank if the range is only one number. / NUMBER ONLY'),
    )
    # this field depends on the site field
    # -> you can only enter deliveries that 
    # -> share the same site as the field site
    delivery = DynamicModelChoiceField(
        label=_('Delivery'),
        queryset=PhoneDelivery.objects.all(),
        required=False,
        help_text=_('Specify how this range is delivered.'),
        query_params={
            'site_id': '$site'
        }
    )
    fieldsets = (
        FieldSet(
            'site',
            name=_('Location')
        ),
        FieldSet(
            'start', 'end', 'delivery',
            name=_('DIDs')
        )
    )

    class Meta:
        model = PhoneDID
        fields = ('site', 'start', 'end', 'delivery')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if 'tags' in self.fields:
            del self.fields['tags']

    def clean(self):

        super().clean()


class PhoneDIDBulkImportForm(NetBoxModelImportForm):
    delivery = CSVModelChoiceField(
        queryset=PhoneDelivery.objects.all(),
        to_field_name='id',
        required=False,
    )
    site = CSVModelChoiceField(
        queryset=Site.objects.all(),
        to_field_name='slug',
        required=True,
    )
    start = forms.IntegerField(
        label=_('Start number'),
        help_text='E164 format',
        required=True
    )
    end = forms.IntegerField(
        label=_('End number'),
        help_text='E164 format - can be left blank if the range is only one number.',
        required=False
    )

    class Meta:
        model = PhoneDID
        fields = ['delivery', 'site', 'start', 'end']

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if 'tags' in self.fields:
            del self.fields['tags']

    def clean(self):
        super().clean()
        if not self.cleaned_data.get('end'):
            self.cleaned_data['end'] = self.cleaned_data['start']



import django_tables2 as tables
from django.utils.translation import gettext_lazy as _

from netbox.tables import NetBoxTable, ChoiceFieldColumn, columns
from tenancy.tables import ContactsColumnMixin

from ..models import PhoneMaintainer


__all__ = (
    'PhoneMaintainerTable',
)

class PhoneMaintainerTable(NetBoxTable, ContactsColumnMixin):
    '''
    table for all Phone Deliveries
    '''
    name = tables.Column(
        verbose_name=_('Name'),
        linkify=True
    )
    status = ChoiceFieldColumn(
        linkify=True
    )
    description = tables.Column()
    comments = columns.MarkdownColumn()

    class Meta(NetBoxTable.Meta):
        model = PhoneMaintainer
        fields = ('pk', 'id', 'actions', 'name', 'status', 'time_zone', 'contacts', 
                    'physical_address', 'shipping_address', 'latitude', 'longitude', 'created', 'last_updated', 'description', 'comments', )
        default_columns = ('pk', 'name', 'status', 'description',)



import django_tables2 as tables
from django.utils.translation import gettext_lazy as _

from netbox.tables import NetBoxTable
from ..models import PhoneDID
from ..utils import format_number


__all__ = (
    'PhoneDIDTable',
)


class PhoneDIDTable(NetBoxTable):
    '''
    table for all DID List
    '''
    delivery = tables.Column(
        verbose_name=_('Delivery'), linkify=True
    )
    site = tables.Column(
        verbose_name=_('Site'), linkify=True
    )
    start = tables.Column(
        verbose_name=_('Start number'), linkify=True,
    )
    end = tables.Column(
        verbose_name=_('End number'), linkify=True,
    )

    class Meta(NetBoxTable.Meta):
        model = PhoneDID
        fields = ('actions', 'pk', 'id', 'start', 'end', 'delivery', 'site', 'created', 'last_updated')
        default_columns = ('start', 'end', 'delivery')

    def render_start(self, record):
        return format_number(record.start)

    def render_end(self, record):
        return format_number(record.end)



from netbox.api.routers import NetBoxRouter

from .views import *


router = NetBoxRouter()

router.register('phone-deliveries', PhoneDeliveryViewSet)
router.register('phone-dids', PhoneDIDViewSet)
router.register('phone-infos', PhoneInfoViewSet)
router.register('phone-maintainers', PhoneMaintainerViewSet)

urlpatterns = router.urls



from django.db.models import Prefetch
from django.utils.translation import gettext_lazy as _
from rest_framework import serializers

from timezone_field.rest_framework import TimeZoneSerializerField

from netbox.api.fields import ChoiceField
from netbox.api.serializers import NetBoxModelSerializer
from dcim.api.serializers import SiteSerializer
from circuits.api.serializers import ProviderSerializer
from dcim.models import Site

from ..models import *


__all__ = (
    'PhoneDeliverySerializer',
    'PhoneDIDSerializer',
    'PhoneInfoSerializer',
    'PhoneMaintainerSerializer',
)


# Briefs Serializers
# -> | for addditional infos
#    | without modifying original


class SiteBriefSerializer(serializers.ModelSerializer):
    url = serializers.HyperlinkedIdentityField(view_name='dcim-api:site-detail')
    
    class Meta:
        model = Site
        fields = ('id', 'url', 'slug', 'name', 'description')


#_______________________________
# Phone Maintainer


class PhoneMaintainerSerializer(NetBoxModelSerializer):
    url = serializers.HyperlinkedIdentityField(
        view_name='plugins-api:sop_phone-api:phonemaintainer-detail'
    )
    status = ChoiceField(
        choices=PhoneMaintainerStatusChoice
    )
    site = serializers.SerializerMethodField()
    time_zone = TimeZoneSerializerField(required=False, allow_null=True)

    class Meta:
        model = PhoneMaintainer
        fields = (
            'id', 'url', 'slug', 'display', 'name', 'status',
            'time_zone', 'physical_address', 'shipping_address', 'latitude', 'longitude',
            'description', 'created', 'last_updated', 'site',
        )
        brief_fields = ('id', 'url', 'slug', 'name', 'description')

    def get_site(self, obj):
        phone_infos = PhoneInfo.objects.filter(maintainer=obj).prefetch_related(
            Prefetch('site', queryset=Site.objects.all())
        )
        site = [pi.site for pi in phone_infos if pi.site]
        return SiteBriefSerializer(site, many=True, context=self.context).data


#_______________________________
# Phone Info (Informations)


class PhoneInfoSerializer(NetBoxModelSerializer):
    url = serializers.HyperlinkedIdentityField(
        view_name='plugins-api:sop_phone-api:phoneinfo-detail'
    )
    site = serializers.SerializerMethodField()
    maintainer = serializers.SerializerMethodField()

    class Meta:
        model = PhoneInfo
        fields = ('id', 'url', 'display', 'site', 'maintainer')

    def get_site(self, obj):
        if not obj.site:
            return None
        return SiteSerializer(obj.site, nested=True, many=False, context=self.context).data

    def get_maintainer(self, obj):
        if not obj.maintainer:
            return None
        maintainer_id = PhoneMaintainer.objects.filter(pk=obj.maintainer.id)
        return PhoneMaintainerSerializer(maintainer_id, nested=True, many=True, context=self.context).data


#_______________________________
# Phone DID (DIDs)


class PhoneDIDSerializer(NetBoxModelSerializer):
    url = serializers.HyperlinkedIdentityField(
        view_name='plugins-api:sop_phone-api:phonedid-detail'
    )
    delivery = serializers.SerializerMethodField(read_only=True)
    site = serializers.SerializerMethodField()

    class Meta:
        model = PhoneDID
        fields = ('id', 'url', 'site', 'delivery', 'start', 'end')

    def get_site(self, obj):
        if not obj.site:
            return None
        return SiteSerializer(obj.site, nested=True, many=False, context=self.context).data

    def get_delivery(self, obj):
        if not obj.delivery:
            return None
        deliv = PhoneDelivery.objects.filter(pk=obj.delivery.id)
        return PhoneDeliverySerializer(deliv, nested=True, many=True, context=self.context).data


#_______________________________
# Phone Delivery


class PhoneDeliverySerializer(NetBoxModelSerializer):
    url = serializers.HyperlinkedIdentityField(
        view_name='plugins-api:sop_phone-api:phonedelivery-detail'
    )
    provider = serializers.SerializerMethodField()
    site = serializers.SerializerMethodField()

    class Meta:
        model = PhoneDelivery
        fields = ('id', 'url', 'display', 'site', 'delivery', 'provider',
            'channel_count', 'status', 'ndi', 'dto',
        )
        brief_fields = ('id', 'url', 'display', 'provider', 'delivery',)

    def get_provider(self, obj):
        if not obj.provider:
            return None
        return ProviderSerializer(obj.provider, nested=True, many=False, context=self.context).data


    def get_site(self, obj):
        if not obj.site:
            return None
        return SiteSerializer(obj.site, nested=True, many=False, context=self.context).data



from django.contrib.auth.mixins import PermissionRequiredMixin
from django.utils.translation import gettext_lazy as _
from django.shortcuts import get_object_or_404

from utilities.views import GetRelatedModelsMixin, register_model_view

from netbox.views import generic
from dcim.models import Site

from ..forms.phone_delivery import *
from ..tables.phone_delivery import *
from ..tables.phone_did import *
from ..filtersets import PhoneDeliveryFilterSet
from ..models import *
from ..utils import count_all_did, format_number


__all__ =  (
    'PhoneDeliveryEditView',
    'PhoneDeliveryDetailView',
    'PhoneDeliveryDeleteView',
    'PhoneDeliveryBulkEditView',
    'PhoneDeliveryDeleteView',
    'PhoneDeliveryListView',
    'PhoneDeliverySiteView'
)


class PhoneDeliveryListView(generic.ObjectListView):
    queryset = PhoneDelivery.objects.all()
    table = PhoneDeliveryTable
    filterset = PhoneDeliveryFilterSet
    filterset_form = PhoneDeliveryFilterForm


class PhoneDeliveryBulkEditView(generic.BulkEditView):
    queryset = PhoneDelivery.objects.all()
    table = PhoneDeliveryTable
    form = PhoneDeliveryBulkEditForm
    filterset = PhoneDeliveryFilterSet


class PhoneDeliveryBulkDeleteView(generic.BulkDeleteView):
    queryset = PhoneDelivery.objects.all()
    table = PhoneDeliveryTable
    filterset = PhoneDeliveryFilterSet


class PhoneDeliveryDetailView(generic.ObjectView, PermissionRequiredMixin, GetRelatedModelsMixin):
    '''
    returns the Phone Delivery detail page with context
    '''
    queryset = PhoneDelivery.objects.all()

    def get_extra_context(self, request, instance) -> dict:
        context: dict = {}

        did = PhoneDID.objects.filter(delivery=instance)

        try:
            site_info = PhoneInfo.objects.filter(site=instance.site.id)
            context['maintainer'] = site_info.first().maintainer
        except:pass
        if instance.ndi:
            context['ndi'] = format_number(instance.ndi)
        if instance.dto:
            context['dto'] = format_number(instance.dto)
        context['did_range'] = PhoneDID
        context['num_range'] = did.count()
        context['num_did'] = count_all_did(did).__int__()
        context['related_models'] = self.get_related_models(
            request, instance,
        )
        return context


class PhoneDeliveryEditView(generic.ObjectEditView):
    '''
    creates anew Phone Delivery instance
    '''
    queryset = PhoneDelivery.objects.all()
    form = PhoneDeliveryForm


class PhoneDeliveryDeleteView(generic.ObjectDeleteView, PermissionRequiredMixin):
    '''
    deletes a Phone Delivery object
    '''
    queryset = PhoneDelivery.objects.all()


class PhoneDeliverySiteView(generic.ObjectEditView):
    '''
    adds a site automatically to the Phone Delivery
    '''
    queryset = PhoneDelivery.objects.all()
    form = PhoneDeliveryForm

    def get_object(self, **kwargs):
        return self.queryset.model(site=get_object_or_404(Site, pk=kwargs['pk']))

    def alter_object(self, obj, request, args, kwargs):
        pk = kwargs.get('pk')
        site = get_object_or_404(Site, pk=pk)
        obj = self.queryset.model
        return obj(site=site)

    def get(self, request, *args, **kwargs): 
        '''
        get request handler
        '''
        response = super().get(request, *args, **kwargs)
        return response



from django.db import models
from django.urls import reverse
from django.db.models import Transform
from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _

from timezone_field import TimeZoneField

from netbox.models import NetBoxModel, PrimaryModel
from netbox.models.features import ContactsMixin
from utilities.choices import ChoiceSet
from circuits.models import Provider
from dcim.models import Site

from .validators import PhoneValidator, PhoneMaintainerValidator
from .utils import format_number


__all__ = (
    'PhoneDID',
    'LogValue',
    'PhoneInfo',
    'FloorValue',
    'PhoneDelivery',
    'AbsoluteValue',
    'PhoneMaintainer',
    'PhoneBoolChoices',
    'PhoneMaintainerStatusChoice',
    'PhoneDeliveryStatusChoices',
)


class PhoneMaintainerStatusChoice(ChoiceSet):

    CHOICES = (
        ('active', _('Active'), 'green'),
        ('retired', _('Retired'), 'red'),
        ('unknown', _('Unknown'), 'gray'),
    )


class PhoneDeliveryStatusChoices(ChoiceSet):

    CHOICES = (
        ('active', _('Active'), 'green'),
        ('planned', _('Planned'), 'cyan'),
        ('staging', _('Staging'), 'blue'),
        ('retired', _('Retired'), 'red'),
        ('unknown', _('Unknown'), 'gray'),
    )


class PhoneBoolChoices(ChoiceSet):

    CHOICES = (
        ('unknown', _('Unknown'), 'gray'),
        ('true', _('True'), 'green'),
        ('false', _('False'), 'red'),
    )


class AbsoluteValue(Transform):
    lookup_name = "abs"
    function = "ABS"
    bilateral = True

class FloorValue(Transform):
    lookup_name = "floor"
    function = "FLOOR"
    bilateral = True

class LogValue(Transform):
    lookup_name = "log"
    function = "LOG"
    bilateral = True


models.IntegerField.register_lookup(AbsoluteValue)
models.IntegerField.register_lookup(FloorValue)
models.IntegerField.register_lookup(LogValue)


class BiAbsoluteValue(Transform):
    lookup_name = "biabs"
    function = "ABS"
    bilateral = True

class BiFloorValue(Transform):
    lookup_name = "bifloor"
    function = "FLOOR"
    bilateral = True

class BiLogValue(Transform):
    lookup_name = "bilog"
    function = "LOG"
    bilateral = True

models.IntegerField.register_lookup(BiAbsoluteValue)
models.IntegerField.register_lookup(BiFloorValue)
models.IntegerField.register_lookup(BiLogValue)

class PhoneMaintainer(PrimaryModel, ContactsMixin):
    name = models.CharField(
        verbose_name=_('Maintainer'),
        unique=True,
    )
    slug = models.SlugField(
        verbose_name=_('slug'),
        max_length=100,
        unique=True,
        blank=True,
    )
    status = models.CharField(
        max_length=30,
        choices=PhoneMaintainerStatusChoice,
        default="Unknown",
        verbose_name=_('Status')
    )
    physical_address = models.CharField(
        verbose_name=_('Physical address'),
        max_length=200,
        null=True,
        blank=True,
        help_text=_('Physical location of the maintainer')
    )
    shipping_address = models.CharField(
        verbose_name=_('Shipping address'),
        max_length=200,
        null=True,
        blank=True,
        help_text=_('If different from the physical address')
    )
    time_zone = TimeZoneField(
        null=True,
        blank=True
    )
    latitude = models.DecimalField(
        verbose_name=_('Latitude'),
        max_digits=8,
        decimal_places=6,
        blank=True,
        null=True,
        help_text=_('GPS coordinate in decimal format (xx.yyyyyy)')
    )
    longitude = models.DecimalField(
        verbose_name=_('Longitude'),
        max_digits=9,
        decimal_places=6,
        blank=True,
        null=True,
        help_text=_('GPS coordinate in decimal format (xx.yyyyyy)')
    )

    def __str__(self) -> str:
        return f'{self.name}'

    def get_absolute_url(self) -> str:
        return reverse('plugins:sop_phone:phonemaintainer_detail', args=[self.pk])

    def get_status_color(self) -> str:
        return PhoneDeliveryStatusChoices.colors.get(self.status)

    class Meta(NetBoxModel.Meta):
        verbose_name = _('Phone Maintainer')
        verbose_name_plural = _('Phone Maintainers')
        constraints = (
            models.UniqueConstraint(
                fields=('name',),
                name='%(app_label)s_%(class)s_unique_name',
                violation_error_message=_("Maintainer name must be unique.")
            ),
            models.UniqueConstraint(
                fields=('slug',),
                name='%(app_label)s_%(class)s_unique_slug',
                violation_error_message=_("Maintainer slug must be unique.")
            )
        )

    def fail(self, message, field=None):
        if field is not None:
            raise ValidationError({field: message})
        raise ValidationError(message)

    def clean(self):
        super().clean()
        if self.physical_address:
            PhoneMaintainerValidator.check_address(self.physical_address, 'physical_address')
        if self.shipping_address:
            PhoneMaintainerValidator.check_address(self.shipping_address, 'shipping_address')
 

class PhoneInfo(NetBoxModel):
    site = models.OneToOneField(
        Site,
        on_delete=models.CASCADE,
        verbose_name=_('Site'),
    )
    maintainer = models.ForeignKey(
        PhoneMaintainer,
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        verbose_name=_('Maintainer'),
    )

    def __str__(self) -> str:
        return f'{self.site} phone maintainer'

    def get_absolute_url(self) -> str:
        return reverse('plugins:sop_phone:phoneinfo_detail', args=[self.pk])

    class Meta(NetBoxModel.Meta):
        verbose_name = _('Information')
        verbose_name_plural = _('Informations')
        constraints = [
            models.UniqueConstraint(
                fields=['site',],
                name='%(app_label)s_%(class)s_unique_site',
                violation_error_message=_("Site must be unique.")
            )
        ]


class PhoneDelivery(NetBoxModel):
    delivery = models.CharField(
        verbose_name=_('Delivery'),
        null=True,
        blank=True,
    )
    provider = models.ForeignKey(
        Provider,
        on_delete=models.RESTRICT,
        verbose_name=_('Provider'),
        null=True,
        blank=True,
    )
    site = models.ForeignKey(
        Site,
        on_delete=models.RESTRICT,
        verbose_name=_('Site'),
    )
    channel_count = models.PositiveBigIntegerField(
        verbose_name=_('Channel Count'),
        null=True,
        blank=True,
    )
    status = models.CharField(
        max_length=30,
        choices=PhoneDeliveryStatusChoices,
        verbose_name=_('Status'),
    )
    ndi = models.PositiveBigIntegerField(
        verbose_name=_('NDI'),
        null=True,
        blank=True,
    )
    dto = models.PositiveBigIntegerField(
        verbose_name=_('DTO'),
        null=True,
        blank=True,
    )
    description = models.CharField(
        verbose_name=_('description'),
        max_length=200,
        blank=True
    )
    comments = models.TextField(
        verbose_name=_('comments'),
        blank=True
    )

    def get_absolute_url(self) -> str:
        return reverse('plugins:sop_phone:phonedelivery_detail', args=[self.pk])

    def get_status_color(self) -> str:
        return PhoneDeliveryStatusChoices.colors.get(self.status)

    def __str__(self) -> str:
        delivery:str = self.delivery if hasattr(self, 'delivery') else 'Unknown delivery'
        provider:str = self.provider if hasattr(self, 'provider') else 'Unknown provider'
        return f'{delivery} / {provider}'

    def clean(self):
        super().clean()
        if hasattr(self, 'delivery') and self.delivery and self.site:
            if PhoneDelivery.objects.exclude(pk=self.pk).filter(site=self.site, delivery=self.delivery).exists():
                raise ValidationError({
                    'delivery': _(f'A "{self.delivery}" delivery method already exists for this site.')
                })

        if self.ndi:
            PhoneValidator.check_number('ndi', self.ndi)
            if PhoneDelivery.objects.exclude(pk=self.pk).filter(ndi=self.ndi).exists():
                raise ValidationError({
                    'ndi': _(f'The MBN/NDI {self.ndi} already exists on another delivery.')
                })
            if PhoneDelivery.objects.filter(pk=self.pk, dto=self.ndi):
                raise ValidationError({
                    'ndi': _(f'The MBN/NDI {self.ndi} cannot be the DTO of its own delivery.')
                    })
            lndi = len(str(self.ndi))
            for rng in PhoneDID.objects.exclude(site=self.site):
                if len(str(rng.start)) == lndi:
                    if self.ndi <= rng.end and rng.start <= self.ndi:
                        raise ValidationError({
                            'ndi': _(f'The MBN/NDI {self.ndi} overlaps DID range {rng.start} -> {rng.end}')
                        })
        if self.dto:
            PhoneValidator.check_number('dto', self.dto)
            if self.ndi == self.dto:
                raise ValidationError({
                    'dto': _(f'The DTO {self.dto} cannot be the MBN/NDI of its own delivery.')
                    })
            current = PhoneDelivery.objects.filter(pk=self.pk)
            if current.exists and PhoneDID.objects.filter(delivery=current.first()):
                ldto = len(str(self.dto))
                for rng in PhoneDID.objects.filter(delivery=current.first()):
                    if len(str(rng.start)) == ldto:
                        if self.dto <= rng.end and rng.start <= self.dto:
                            raise ValidationError({
                                'dto': _(f'The DTO {self.dto} overlaps DID range {rng.start} -> {rng.end}')
                                })

    class Meta(NetBoxModel.Meta):
        verbose_name = _('Phone Delivery')
        verbose_name_plural = _('Phone Deliveries')
        constraints = (
            models.UniqueConstraint(
                fields=('ndi',),
                name='%(app_label)s_%(class)s_unique_ndi',
                violation_error_message=_("NDI must be unique.")
            ),
            models.UniqueConstraint(
                fields=('delivery', 'site'),
                name='%(app_label)s_%(class)s_unique_delivery_method_site',
                violation_error_message=_("Delivery method must be unique in a site.")
            )
        )


class PhoneDID(NetBoxModel):
    delivery = models.ForeignKey(
        PhoneDelivery,
        on_delete=models.SET_NULL,
        verbose_name=_('Delivery'),
        null=True,
        blank=True,
    )
    site = models.ForeignKey(
        Site,
        on_delete=models.CASCADE,
        verbose_name=_('Site'),
        null=False,
        blank=True,
    )
    start = models.PositiveBigIntegerField(
        unique=False,
        verbose_name=_('Start number'),
        null=False,
        blank=True,
    )
    end = models.PositiveBigIntegerField(
        unique=False,
        verbose_name=_('End number'),
        null=False,
        blank=True,
    )

    def __str__(self) -> str:
        if not self.start and not self.end:
            return f'No DID'
        # \u00A0 is unicode for ' '
        return f'{format_number(self.start)}\u00A0\u00A0\u00A0\u00A0>>\u00A0\u00A0\u00A0\u00A0{format_number(self.end)}'

    def get_absolute_url(self) -> str:
        return reverse('plugins:sop_phone:phonedid_detail', args=[self.pk])

    def clean(self):
        super().clean()

        PhoneValidator.check_delivery_site(self.delivery, self.site)

        PhoneValidator.check_number('start', self.start)
        if self.end is None:
            self.end=self.start
        else :
            PhoneValidator.check_number('end', self.end)
            PhoneValidator.check_start_end(self.start, self.end)  

        if PhoneDelivery.objects.exclude(site=self.site).filter(ndi=self.start).exists():
            raise ValidationError({
                'start': _(f'Start {self.start} overlaps a delivery MBN/NDI.')
                })
        if PhoneDelivery.objects.exclude(site=self.site).filter(ndi=self.end).exists():
            raise ValidationError({
                'end': _(f'End {self.end} overlaps a delivery MBN/NDI')
                })

        # check if the current range overlaps its own DTO
        PhoneValidator.check_delivery_overlaps(self.delivery, self.start, self.end)
        # check if self.end or self.start overlaps an existing DID range
        lnum=len(str(self.start))
        for ndi in PhoneDelivery.objects.exclude(site=self.site).values_list('ndi', flat=True):
            #only compare if numbers are comaprable (have the same length)
            if len(str(ndi)) == lnum:
                # check if number overlap
                if self.start <= ndi and self.end >= ndi:
                    raise ValidationError({
                        'start': _(f'This range {self.start} -> {self.end} overlaps a delivery MBN/NDI')
                    })

        for rng in PhoneDID.objects.exclude(pk=self.pk):
            #only compare if numbers are comaprable (have the same length)
            if len(str(rng.start))==lnum:
                # check if number overlap
                if self.start <= rng.end and rng.start <= self.end:
                    raise ValidationError({
                        'start': _(f'This range {self.start} -> {self.end} overlaps range {rng.start} -> {rng.end}.')
                    })

    def save(self, *args, **kwargs):
        if not self.end or self.end == 0:
            self.end = self.start
        super().save(*args, **kwargs)

    class Meta(NetBoxModel.Meta):
        ordering = ('start',)
        verbose_name = _('DID Range')
        verbose_name_plural = _('DIDs')

        constraints = (
            models.UniqueConstraint(
                fields=('start',),
                name='%(app_label)s_%(class)s_unique_start_number',
                violation_error_message=_("Start number must be unique.")
            ),
            models.UniqueConstraint(
                fields=('end',),
                name='%(app_label)s_%(class)s_unique_end_number',
                violation_error_message=_("End number must be unique.")
            ),
            models.CheckConstraint(
                check=models.Q(end__gte=models.F('start')) & \
                    models.Q(start__biabs__bilog__bifloor=models.F("end")),
                name='%(app_label)s_%(class)s_end_greater_than_start',
                violation_error_message=_("End number must be greater than or equal to start number.")
            )
        )



from django.urls import path

from netbox.views.generic import ObjectChangeLogView, ObjectJournalView

from .views import phone_maintainer as vm
from .views import phone_delivery as vd
from .views import phone_did as did
from .views import phone_info as pi

'''🡣 DO NOT REMOVE 🡣'''
from .views import tab_view
'''🡩 DO NOT REMOVE 🡩'''

from .models import *


app_name = 'sop_phone'


urlpatterns = [

    # phone maintainer
    path('phone-maintainer/', vm.PhoneMaintainerListView.as_view(), name='phonemaintainer_list'),
    path('phone-maintainer/<int:pk>', vm.PhoneMaintainerView.as_view(), name='phonemaintainer_detail'),
    path('phone-maintainer/add', vm.PhoneMaintainerEditView.as_view(), name='phonemaintainer_add'),
    path('phone-maintainer/edit', vm.PhoneMaintainerBulkEditView.as_view(), name='phonemaintainer_bulk_edit'),
    path('phone-maintainer/import', vm.PhoneMaintainerBulkImportView.as_view(), name='phonemaintainer_import'),
    path('phone-maintainer/delete/', vm.PhoneMaintainerBulkDeleteView.as_view(), name='phonemaintainer_bulk_delete'),
    path('phone-maintainer/<int:pk>/contacts/', vm.PhoneMaintainerContactsView.as_view(), name='phonemaintainer_contacts'),
    path('phone-maintainer/edit/<int:pk>', vm.PhoneMaintainerEditView.as_view(), name='phonemaintainer_edit'),
    path('phone-maintainer/delete/<int:pk>', vm.PhoneMaintainerDeleteView.as_view(), name='phonemaintainer_delete'),
    path('phone-maintainer/changelog/<int:pk>', ObjectChangeLogView.as_view(), name='phonemaintainer_changelog', kwargs={'model': PhoneMaintainer}),
    path('phone-maintainer/journal/<int:pk>', ObjectJournalView.as_view(), name='phonemaintainer_journal', kwargs={'model': PhoneMaintainer}),

    # phone delivery
    path('phone-delivery/', vd.PhoneDeliveryListView.as_view(), name='phonedelivery_list'),
    path('phone-delivery/<int:pk>', vd.PhoneDeliveryDetailView.as_view(), name='phonedelivery_detail'),
    path('phone-delivery/add', vd.PhoneDeliveryEditView.as_view(), name='phonedelivery_add'),
    path('phone-delivery/add_site/<int:pk>', vd.PhoneDeliverySiteView.as_view(), name='phonedelivery_site_add'),
    path('phone-delivery/edit', vd.PhoneDeliveryBulkEditView.as_view(), name='phonedelivery_bulk_edit'),
    path('phone-delivery/delete', vd.PhoneDeliveryBulkDeleteView.as_view(), name='phonedelivery_bulk_delete'),
    path('phone-delivery/edit/<int:pk>', vd.PhoneDeliveryEditView.as_view(), name='phonedelivery_edit'),
    path('phone-delivery/delete/<int:pk>', vd.PhoneDeliveryDeleteView.as_view(), name='phonedelivery_delete'),
    path('phone-delivery/changelog/<int:pk>', ObjectChangeLogView.as_view(), name='phonedelivery_changelog', kwargs={'model': PhoneDelivery}),
    path('phone-delivery/journal/<int:pk>', ObjectJournalView.as_view(), name='phonedelivery_journal', kwargs={'model': PhoneDelivery}),

    # phone did
    path('phone-did/', did.PhoneDIDListView.as_view(), name='phonedid_list'),
    path('phone-did/<int:pk>', did.PhoneDIDDetailView.as_view(), name='phonedid_detail'),
    path('phone-did/add', did.PhoneDIDEditView.as_view(), name='phonedid_add'),
    path('phone-did/add_site/<int:pk>', did.PhoneDIDAddSiteView.as_view(), name='phonedid_site_add'),
    path('phone-did/edit/', did.PhoneDIDBulkEditView.as_view(),  name='phonedid_bulk_edit'),
    path('phone-did/import/', did.PhoneDIDBulkImportView.as_view(), name='phonedid_import'),
    path('phone-did/delete/', did.PhoneDIDBulkDeleteView.as_view(), name='phonedid_bulk_delete'),
    path('phone-did/edit/<int:pk>', did.PhoneDIDEditView.as_view(), name='phonedid_edit'),
    path('phone-did/delete/<int:pk>', did.PhoneDIDDeleteView.as_view(), name='phonedid_delete'),
    path('phone-did/changelog/<int:pk>', ObjectChangeLogView.as_view(), name='phonedid_changelog', kwargs={'model': PhoneDID}),
    path('phone-did/journal/<int:pk>', ObjectJournalView.as_view(), name='phonedid_journal', kwargs={'model': PhoneDID}),
 
    # phone info
    path('phone-info/', pi.PhoneInfoListView.as_view(), name='phoneinfo_list'),
    path('phone-info/<int:pk>', pi.PhoneInfoDetailView.as_view(), name='phoneinfo_detail'),
    path('phone-info/add/', pi.PhoneInfoAddView.as_view(), name='phoneinfo_add'),
    path('phone-info/add/<int:pk>', pi.PhoneInfoAddView.as_view(), name='phoneinfo_add'),
    path('phone-info/edit/', pi.PhoneInfoBulkEditView.as_view(), name='phoneinfo_bulk_edit'),
    path('phone-info/delete/', pi.PhoneInfoBulkDeleteView.as_view(), name='phoneinfo_bulk_delete'),
    path('phone-info/edit/<int:pk>', pi.PhoneInfoEditView.as_view(), name='phoneinfo_edit'),
    path('phone-info/delete/<int:pk>', pi.PhoneInfoDeleteView.as_view(), name='phoneinfo_delete'),
    path('phone-info/changelog/<int:pk>', ObjectChangeLogView.as_view(), name='phoneinfo_changelog', kwargs={'model': PhoneInfo}),
    path('phone-info/journal/<int:pk>', ObjectJournalView.as_view(), name='phoneinfo_journal', kwargs={'model': PhoneInfo}),

]



from django import forms
from django.utils.translation import gettext_lazy as _
from django.utils.safestring import mark_safe

from timezone_field import TimeZoneFormField

from utilities.forms import add_blank_choice
from tenancy.forms import ContactModelFilterForm
from utilities.forms.fields import (
    CommentField, SlugField, CSVChoiceField, DynamicModelChoiceField
)
from utilities.forms.rendering import FieldSet
from netbox.forms import (
    NetBoxModelFilterSetForm, NetBoxModelForm, NetBoxModelBulkEditForm,
    NetBoxModelImportForm
)
from dcim.models import Site, Region, SiteGroup

from ..models import *


__all__ = (
    'PhoneMaintainerForm',
    'PhoneMaintainerFilterForm',
    'PhoneMaintainerBulkEditForm',
    'PhoneMaintainerBulkImportForm',
)


class PhoneMaintainerFilterForm(NetBoxModelFilterSetForm, ContactModelFilterForm):
    model = PhoneMaintainer

    site_id = DynamicModelChoiceField(
        queryset=Site.objects.all(),
        required=False,
        label=_('Site')
    )
    region_id = DynamicModelChoiceField(
        queryset=Region.objects.all(),
        required=False,
        label=_('Region')
    )
    group_id = DynamicModelChoiceField(
        queryset=SiteGroup.objects.all(),
        required=False,
        label=_('Site group')
    )
    status = forms.ChoiceField(
        choices=add_blank_choice(PhoneMaintainerStatusChoice),
        initial=None,
        required=False,
        label=_('Status'),
    )

    fieldsets = (
        FieldSet(
            'region_id', 'group_id', 'site_id',
            name=_('Location')
        ),
        FieldSet(
            'contact', 'contact_role', 'contact_group',
            name=_('Contacts')
        ),
        FieldSet(
            'name', 'slug', 'status',
            name=_('Attributes')
        )
    )
    



class PhoneMaintainerBulkEditForm(NetBoxModelBulkEditForm):
    model = PhoneMaintainer

    status = forms.ChoiceField(
        choices=PhoneMaintainerStatusChoice,
        required=True,
    )

    class Meta:
        fields = ('status', )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if 'add_tags' in self.fields:
            del self.fields['add_tags']
        if 'remove_tags' in self.fields:
            del self.fields['remove_tags']


class PhoneMaintainerForm(NetBoxModelForm):
    name = forms.CharField(label=_('Name'))
    slug = SlugField()
    status = forms.ChoiceField(
        choices=PhoneMaintainerStatusChoice,
        required=True
    )
    time_zone = TimeZoneFormField(
        label=_('Time zone'),
        choices=add_blank_choice(TimeZoneFormField().choices),
        required=False
    )
    comments = CommentField()

    fieldsets = (
        FieldSet(
            'name', 'slug', 'status', 'time_zone', 'description',
            name=_('Maintainer')
        ),
        FieldSet(
            'physical_address', 'shipping_address', 'latitude', 'longitude',
            name=_('Contact Info')
        )
    )

    class Meta:
        model = PhoneMaintainer
        fields = (
            'name', 'slug', 'status', 'time_zone', 'description',
            'physical_address', 'shipping_address', 'latitude', 'longitude',
            'comments'
        )
        widgets = {
            'physical_address': forms.Textarea(
                attrs={
                    'rows': 3,
                }
            ),
            'shipping_address': forms.Textarea(
                attrs={
                    'rows': 3,
                }
            ),
        }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if 'tags' in self.fields:
            del self.fields['tags']


class PhoneMaintainerBulkImportForm(NetBoxModelImportForm):
    status = CSVChoiceField(
        choices=PhoneMaintainerStatusChoice,
        required=True,
    )
    slug = SlugField(required=True)

    class Meta:
        model = PhoneMaintainer
        fields = [
            'name', 'slug', 'status', 'time_zone', 'description',
            'physical_address', 'shipping_address', 'latitude', 'longitude',
            'comments'
        ]
        help_texts = {
            'time_zone': mark_safe(
                '{} (<a href="https://en.wikipedia.org/wiki/List_of_tz_database_time_zones">{}</a>)'.format(
                    _('Time zone'), _('available options')
                )
            )
        }

    def clean(self):
        super().clean()



from django import forms
from django.utils.translation import gettext_lazy as _

from circuits.models import Provider
from utilities.forms import add_blank_choice
from utilities.forms.fields import SlugField
from dcim.models import Site, Region, SiteGroup
from utilities.forms.fields import DynamicModelChoiceField, CommentField
from netbox.forms import NetBoxModelFilterSetForm, NetBoxModelForm, NetBoxModelBulkEditForm
from utilities.forms.rendering import FieldSet

from ..models import *


__all__ = (
    'PhoneDeliveryForm',
    'PhoneDeliveryFilterForm',
    'PhoneDeliveryBulkEditForm',
)


class PhoneDeliveryFilterForm(NetBoxModelFilterSetForm):
    model = PhoneDelivery

    site_id = DynamicModelChoiceField(
        queryset=Site.objects.all(),
        required=False,
        label=_('Site')
    )
    group_id = DynamicModelChoiceField(
        queryset=SiteGroup.objects.all(),
        required=False,
        label=_('Site group')
    )
    region_id = DynamicModelChoiceField(
        queryset=Region.objects.all(),
        required=False,
        label=_('Region')
    )
    maintainer_id = DynamicModelChoiceField(
        queryset=PhoneMaintainer.objects.all(),
        required=False,
        label=_('Maintainer')
    )
    delivery_id = forms.CharField(
        label=_('Delivery Method'),
        required=False
    )
    provider = forms.ModelChoiceField(
        queryset=Provider.objects.all(),
        required=False,
        label=_('Provider')
    )
    status = forms.MultipleChoiceField(
        choices=add_blank_choice(PhoneDeliveryStatusChoices),
        initial=None,
        required=False,
        label=_('Status')
    )
    channel_count = forms.IntegerField(
        required=False,
        label=_('Channel count'),
        help_text=_('G.711 cidec - 96kbps reserved bandwidth per channel / NUMBER ONLY')
    )
    ndi = forms.IntegerField(
        required=False,
        label=_('MBN / NDI'),
        help_text=_("Main Billing Number / Numéro de Désignation d'Installation - E164 format / NUMBER ONLY")
    )
    dto = forms.IntegerField(
        required=False,
        label=_('DTO'),
        help_text=_('E164 format / NUMBER ONLY')
    )

    fieldsets = (
        FieldSet(
            'region_id', 'group_id', 'site_id',
            name=_('Location')
        ),
        FieldSet(
            'maintainer_id',
            name=_('Information')
        ),
        FieldSet(
            'delivery_id', 'provider', 'status',
            'channel_count', 'ndi', 'dto',
            name=_('Attributes')
        )
    )


class PhoneDeliveryBulkEditForm(NetBoxModelBulkEditForm):
    model = PhoneDelivery

    site = forms.ModelChoiceField(
        queryset=Site.objects.all(),
        required=False,
        label=_('Site')
    )
    delivery = forms.CharField(
        required=False,
        label=_('Delivery Method'),
        help_text=_('SIP TRUNK, T0, T2, ...')
    )
    provider = forms.ModelChoiceField(
        queryset=Provider.objects.all(),
        required=False,
        label=_('Provider')
    )
    status = forms.ChoiceField(
        choices=PhoneDeliveryStatusChoices,
        required=False,
        label=_('Status')
    )

    class Meta:
        fields = ('site', 'delivery', 'provider', 'status', )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if 'add_tags' in self.fields:
            del self.fields['add_tags']
        if 'remove_tags' in self.fields:
            del self.fields['remove_tags']


class PhoneDeliveryForm(NetBoxModelForm):
    '''
    creates a form for a Phone Delivery object
    '''
    site = forms.ModelChoiceField(
        required=True,
        queryset=Site.objects.all(),
        label=_('Site')
    )
    delivery = forms.CharField(
        required=True,
        label=_('Delivery Method'),
        help_text=_('SIP TRUNK, T0, T2, ...')
    )
    provider = forms.ModelChoiceField(
        required=True,
        queryset=Provider.objects.all(),
        label=_('Provider')
    )
    channel_count = forms.IntegerField(
        required=False,
        label=_('Channel Count'),
        help_text=_('G.711 cidec - 96kbps reserved bandwidth per channel / NUMBER ONLY')
    )
    status = forms.ChoiceField(
        choices=PhoneDeliveryStatusChoices,
        required=True,
        label=_('Status'),
    )
    ndi = forms.IntegerField(
        required=False,
        label=_('MBN / NDI'),
        help_text=_("Main Billing Number / Numéro de Désignation d'Installation - E164 format / NUMBER ONLY")
    )
    dto = forms.IntegerField(
        required=False,
        label=_('DTO'),
        help_text=_('E164 format / NUMBER ONLY')
    )
    comments = CommentField()

    fieldsets = (
        FieldSet(
            'delivery', 'provider', 'status', 'channel_count', 'ndi', 'dto', 'description',
            name=_('Delivery')
        ),
        FieldSet(
            'site',
            name=_('Location')
        )
    )

    class Meta:
        model = PhoneDelivery
        fields = ('site', 'delivery', 'provider', 'channel_count', 'status', 'ndi', 'dto', 'description', 'comments')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if 'tags' in self.fields:
            del self.fields['tags']



import django_tables2 as tables
from django.utils.translation import gettext_lazy as _

from netbox.tables import NetBoxTable, ChoiceFieldColumn

from ..models import PhoneDelivery
from ..utils import format_number


__all__ = (
    'PhoneDeliveryTable',
)

class PhoneDeliveryTable(NetBoxTable):
    '''
    table for all Phone Deliveries
    '''
    delivery = tables.Column(
        verbose_name=_('Delivery Method'),
        linkify=True
    )
    provider = tables.Column(
        verbose_name=_('Provider'),
        linkify=True
    )
    site = tables.Column(
        verbose_name=_('Site'),
        linkify=True
    )
    channel_count = tables.Column(
        verbose_name=_('Channel Count'),
        linkify=True
    )
    status = ChoiceFieldColumn(linkify=True)
    ndi = tables.Column(
        verbose_name=_('MBN / NDI'),
        linkify=True
    )
    dto = tables.Column(
        verbose_name=_('DTO'),
        linkify=True
    )

    class Meta(NetBoxTable.Meta):
        model = PhoneDelivery
        fields = ('pk', 'id', 'actions', 'delivery', 'provider', 'site',
            'channel_count', 'status', 'ndi', 'dto', 'created', 'last_updated', 'description', 'comments')
        default_columns = ('actions', 'site', 'delivery', 'channel_count', 'status', 'ndi', 'dto', 'description')

    def render_ndi(self, record):
        return format_number(record.ndi)

    def render_dto(self, record):
        return format_number(record.dto)



import django_filters
from django.db.models import Q
from django.utils.translation import gettext_lazy as _ 

from dcim.models import Site, Region, SiteGroup
from netbox.filtersets import NetBoxModelFilterSet
from utilities.filters import MultiValueCharFilter
from tenancy.filtersets import  ContactModelFilterSet

from .models import *
from .validators import number_quicksearch


__all__ = (
    'PhoneDeliveryFilterSet',
    'PhoneInfoFilterSet',
    'PhoneMaintainerFilterSet',
    'PhoneDIDFilterSet',
)


#_________________________
# Phone Delivery Filters

class PhoneDeliveryFilterSet(NetBoxModelFilterSet):
    status = django_filters.MultipleChoiceFilter(
        choices=PhoneDeliveryStatusChoices,
        null_value=None
    )
    site_id = django_filters.ModelMultipleChoiceFilter(
        queryset=Site.objects.all(),
        field_name='site',
        label=_('Site (ID)')
    )
    site_name = django_filters.CharFilter(
        field_name='site__name',
        label=_('Site (name)')
    )
    region_id = django_filters.ModelMultipleChoiceFilter(
        queryset=Region.objects.all(),
        field_name='site__region',
        label=_('Region (ID)')
    )
    group_id = django_filters.ModelMultipleChoiceFilter(
        queryset=SiteGroup.objects.all(),
        field_name='site__group',
        label=_('Site group (ID)')
    )
    maintainer_id = django_filters.ModelMultipleChoiceFilter(
        queryset=PhoneMaintainer.objects.all(),
        method='search_maintainer_id',
        label=_('Maintainer (ID)')
    )
    maintainer_name = django_filters.CharFilter(
        method='search_maintainer_name',
        label=_('Maintainer (name')
    )

    class Meta:
        model = PhoneDelivery
        fields = ('id', 'site', 'delivery', 'provider', 'status', 'channel_count', 'ndi', 'dto')

    def search_maintainer_id(self, queryset, name, value):
        if not value:
            return queryset
        try:
            site_ids = PhoneInfo.objects.filter(maintainer__in=value).values_list('site__id', flat=True )
            return queryset.filter(site__id__in=site_ids)
        except:return queryset

    def search_maintainer_name(self, queryset, name, value):
        if not value:
            return queryset
        try:
            maintainer_ids = PhoneMaintainer.objects.filter(name=value).values_list('id', flat=True)
            site_ids = PhoneInfo.objects.filter(maintainer__in=maintainer_ids).values_list('site__id', flat=True)
            return queryset.filter(site__id__in=site_ids)
        except:return queryset

    def search(self, queryset, name, value):
        if not value.strip():
            return queryset
        return queryset.filter(
            Q(site__name__icontains=value) |
            Q(delivery__icontains=value) |
            Q(provider__name__icontains=value) |
            Q(channel_count__icontains=value) |
            Q(dto__icontains=value) |
            Q(ndi__icontains=value)
        )


#_________________________
# Informations filters

class PhoneInfoFilterSet(NetBoxModelFilterSet):

    site_id = django_filters.ModelMultipleChoiceFilter(
        queryset=Site.objects.all(),
        field_name='site__id',
        method='search_site_id',
        label=_('Site (ID)')
    )
    site_name = django_filters.CharFilter(
        field_name='site__name',
        label=_('Site (name)')
    )
    region_id = django_filters.ModelMultipleChoiceFilter(
        queryset=Region.objects.all(),
        field_name='site__region',
        label=_('Region (ID)')
    )
    group_id = django_filters.ModelMultipleChoiceFilter(
        queryset=SiteGroup.objects.all(),
        field_name='site__group',
        label=_('Site group (ID)')
    )
    maintainer_id = django_filters.ModelMultipleChoiceFilter(
        queryset=PhoneMaintainer.objects.all(),
        field_name='maintainer',
        label=_('Maintainer (ID)')
    )
    maintainer_name = django_filters.CharFilter(
        field_name='maintainer__name',
        label=_('Maintainer (name)')
    )

    class Meta:
        model = PhoneInfo
        fields = ('id', 'site', 'site_id', 'site_name', 'maintainer_id', 'maintainer_name',)

    def search_site_id(self, queryset, name, value):
        if not value:
            return queryset
        return queryset.filter(site__in=value)

    def search(self, queryset, name, value):
        if not value.strip():
            return queryset
        return queryset.filter(
            Q(maintainer__icontains=value) |
            Q(site__icontains=value)
        )


#_________________________
# Maintainers filter

class PhoneMaintainerFilterSet(NetBoxModelFilterSet, ContactModelFilterSet):
    status = django_filters.MultipleChoiceFilter(
        choices=PhoneMaintainerStatusChoice,
        null_value=None
    )
    site_name = django_filters.CharFilter(
        method='search_site_name',
        label=_('Site (name)')
    )
    site_id = django_filters.ModelMultipleChoiceFilter(
        queryset=Site.objects.all(),
        method='search_site_id',
        label=_('Site (ID)')
    )
    region_id = django_filters.ModelMultipleChoiceFilter(
        queryset=Region.objects.all(),
        method='search_region_id',
        label=_('Region (ID)')
    )
    group_id = django_filters.ModelMultipleChoiceFilter(
        queryset=SiteGroup.objects.all(),
        method='search_group_id',
        label=_('Site group (ID)')
    )
    time_zone = MultiValueCharFilter()

    class Meta:
        model = PhoneMaintainer
        fields = ('id', 'name', 'slug', 'status', 'latitude', 'longitude', 'description')

    def search_region_id(self, queryset, name, value):
        if not value:
            return queryset
        try:
            site = Site.objects.filter(region__in=value)
            maintainer_id = PhoneInfo.objects.filter(site__in=site).values_list('maintainer_id', flat=True)
            return queryset.filter(pk__in=maintainer_id)
        except:return queryset

    def search_group_id(self, queryset, name, value):
        if not value:
            return queryset
        try:
            site = Site.objects.filter(group__in=value)
            maintainer_id = PhoneInfo.objects.filter(site__in=site).values_list('maintainer_id', flat=True)
            return queryset.filter(pk__in=maintainer_id)
        except:return queryset

    def search_site_name(self, queryset, name, value):
        if not value:
            return queryset
        try:
            maintainer_id = PhoneInfo.objects.filter(site__in=Site.objects.filter(name=value)).values_list('maintainer_id', flat=True)
            return queryset.filter(pk__in=maintainer_id)
        except:return queryset

    def search_site_id(self, queryset, name, value):
        if not value:
            return queryset
        try:
            maintainer_id = PhoneInfo.objects.filter(site__in=value).values_list('maintainer_id', flat=True)
            return queryset.filter(pk__in=maintainer_id)
        except:return queryset

    def search(self, queryset, name, value):
        if not value.strip():
            return queryset
        return queryset.filter(
            Q(name__icontains=value) |
            Q(slug__icontains=value) |
            Q(physical_address__icontains=value) |
            Q(shipping_address__icontains=value) |
            Q(description__icontains=value) |
            Q(comments__icontains=value)
        )


#_________________________
# DID filters (DIDs)

class PhoneDIDFilterSet(NetBoxModelFilterSet):
    site_id = django_filters.ModelMultipleChoiceFilter(
        queryset=Site.objects.all(),
        field_name='site',
        label=_('Site (ID)')
    )
    site_name = django_filters.CharFilter(
        field_name='site__name',
        label=_('Site (name)')
    )
    maintainer_id = django_filters.ModelMultipleChoiceFilter(
        queryset=PhoneMaintainer.objects.all(),
        field_name='site',
        method='did_maintainer_filter',
        label=_('Maintainer (ID)')
    )
    maintainer_name = django_filters.CharFilter(
        method='did_maintainer_name_filter',
        label=_('Maintainer (name)')
    )
    delivery_id = django_filters.ModelMultipleChoiceFilter(
        queryset=PhoneDelivery.objects.all(),
        field_name='delivery',
        label=_('Delivery (ID)')
    )
    partial_number = django_filters.NumberFilter(
        method='search_partial_number',
        label=_('Partial number')
    )
    region_id = django_filters.ModelMultipleChoiceFilter(
        queryset=Region.objects.all(),
        field_name='site__region',
        label=_('Region')
    )
    group_id = django_filters.ModelMultipleChoiceFilter(
        queryset=SiteGroup.objects.all(),
        field_name='site__group',
        label=_('Site group')
    )

    class Meta:
        model = PhoneDID
        fields = ('id', 'start', 'end', 'site', 'delivery_id')

    def search_partial_number(self, queryset, name, value):
        if not value:
            return queryset

        valid_ids: list[int] = []

        for rng in queryset:
            if number_quicksearch(rng.start, rng.end, str(value)):
                valid_ids.append(rng.id)

        return queryset.filter(id__in=valid_ids)

    def did_maintainer_name_filter(self, queryset, name, value):
        if not value:
            return queryset
        try:
            site_ids = PhoneInfo.objects.filter(maintainer__name=value).values_list('site_id', flat=True)
            return queryset.filter(site_id__in=site_ids)
        except:return queryset

    def did_maintainer_filter(self, queryset, name, value):
        if not value:
            return queryset
        try:
            site_ids = PhoneInfo.objects.filter(maintainer__in=value).values_list('site_id', flat=True)
            return queryset.filter(site_id__in=site_ids)
        except:return queryset

    def search(self, queryset, name, value):
        if not value.strip():
            return queryset
        return queryset.filter(
            Q(start__icontains=value) |
            Q(end__icontains=value) |
            Q(site__name__icontains=value) |
            Q(delivery__delivery__icontains=value) |
            Q(delivery__provider__name__icontains=value)
        )



import phonenumbers


__all__ = (
    'format_number',
    'count_all_did',
    'get_object_or_create',
)


def format_number(number:int) -> str:
    '''
    formats E164 numbers
    and displays a beautiful flag
    depending on their country code
    '''
    def country_code_to_flag(county) -> str:
        '''
        returns the right flag depending on the country code
        '''
        return chr(ord(country[0]) + 127397) + chr(ord(country[1]) + 127397)

    prepare_number:str = f'+{str(number)}'
    parsed_number = phonenumbers.parse(prepare_number)
    country:str = phonenumbers.region_code_for_country_code(parsed_number.country_code)
    flag = country_code_to_flag(country)
    '''
    returns {flag} <spaces> {number}
    '''
    return f'{flag}\u00A0\u00A0\u00A0{phonenumbers.format_number(parsed_number, phonenumbers.PhoneNumberFormat.INTERNATIONAL)}'


def get_object_or_create(model, site) -> object | None:
    '''
    get the model instance or create it
    (for dcim/site extra models)
    '''
    target = model.objects.filter(site=site)

    if target.exists():

        if target.filter(maintainer__isnull=False).exists():
            return target.filter(maintainer__isnull=False).first()

        return target.first()
    
    target = model(site=site)
    target.save()

    return target


class count_all_did:
    """counts all numbers inside a given PhoneDID instance
    
    Args:
        PhoneDID
    Returns:
```python
    __int__(self):
        return self.phone_count)
```
    """
    def __init__(self, phone_did, delivery=None) -> None:
        self.phone_did = phone_did
        self.delivery = delivery
        self.phone_count = self.count()

    def count_range(self, start:int, end:int) -> int:
        count:int = 0

        if start == end:
            return 1
        if start > end:
            return 0
        if start < end:
            count = (end - start) + 1
        return count

    def count_ndi_outside_ranges(self, phone_count) -> int:
        # if no delivery provided, abort
        if not self.delivery:
            return phone_count

        # iterable only in ndi
        for ndi in self.delivery.values_list('ndi', flat=True):
            ndi_in_range:bool = False

            # only count if not in range
            for did in self.phone_did:
                if did.start <= ndi <= did.end:
                    ndi_in_range = True
                    break

            if not ndi_in_range:
                phone_count += 1

        return phone_count

    def count(self) -> int:
        phone_count:int = 0

        # count did sda
        try:
            # try iterable
            for did in self.phone_did:
                phone_count += self.count_range(did.start, did.end)

        except:
            try:
                phone_count += self.count_range(self.phone_did.start, self.phone_did.end)
            except:pass
        
        # if no delivery provided, abort
        if not self.delivery:
            return phone_count

        return self.count_ndi_outside_ranges(phone_count)

    def __int__(self) -> int:
        '''
        ```python
        returns self.phone_count
        ```
        '''
        return self.phone_count




from django.shortcuts import render, get_object_or_404
from django.utils.translation import gettext_lazy as _
from django.shortcuts import redirect
from django.views import View

from utilities.views import register_model_view, ViewTab, ContentTypePermissionRequiredMixin
from utilities.permissions import get_permission_for_model
from netbox.views.generic.mixins import ActionsMixin
from dcim.models import Site

from ..tables.phone_delivery import *
from ..tables.phone_did import *
from ..models import *
from ..utils import count_all_did


'''
accessed for "Phone" site tab View
'''


__all__ = (
    'PhoneSiteTabView',
)


@register_model_view(Site, name="phone")
class PhoneSiteTabView(View, ContentTypePermissionRequiredMixin, ActionsMixin):
    '''
    creates a "phone" tab on the site page
    '''
    tab = ViewTab(
        label="Phone",
        badge=lambda obj: count_all_did(
            PhoneDID.objects.filter(site=obj),
            PhoneDelivery.objects.filter(site=obj)
        ).__int__()
    )
    template_name = "sop_phone/tab/tab.html"

    def get_table(self, table, qs, request):
        table = table(qs, user=request.user)
        if 'pk' in table.base_columns:
            table.columns.show('pk')
        table.configure(request)
        return table

    def get_actions(self, request):
        actions:dict = dict()

        actions['DID'] = self.get_permitted_actions(request.user, PhoneDID)
        actions['Delivery'] = self.get_permitted_actions(request.user, PhoneDelivery)
        actions['Info'] = self.get_permitted_actions(request.user, PhoneInfo)
        return actions

    def get_extra_context(self, request, pk) -> dict:
        '''
        returns all the models and tables needed for the tab
        '''
        context: dict = {}
    
        site = get_object_or_404(Site, pk=pk)
        if PhoneInfo.objects.filter(site=site).exists():
            context['phone_info'] = PhoneInfo.objects.get(site=site)
        context['did'] = PhoneDID
        context['did_table'] = self.get_table(PhoneDIDTable, PhoneDID.objects.filter(delivery__site=site), request)
        context['delivery'] = PhoneDelivery
        context['delivery_table'] = self.get_table(PhoneDeliveryTable, PhoneDelivery.objects.filter(site=site), request)
        context['context'] = context
        context['actions'] = self.get_actions(request)

        '''
        change the devices with the filter you want
        '''
        # context['devices'] = Device.objects.filter(site=site)

        return {'object': site, 'context': context}

    def get(self, request, pk):
        return render(request, self.template_name,
            self.get_extra_context(request, pk))



from netbox.search import SearchIndex, register_search

from .models import *


@register_search
class PhoneDIDSearchIndex(SearchIndex):
    model = PhoneDID
    fields = (
        ('start', 100),
        ('end', 100),
    )


@register_search
class PhoneDeliverySearchIndex(SearchIndex):
    model = PhoneDelivery
    fields = (
        ('delivery', 100),
        ('provider', 100),
        ('site', 500),
    )


@register_search
class PhoneMaintainerSearchIndex(SearchIndex):
    model = PhoneMaintainer
    fields = (
        ('name', 100),
        ('slug', 100),
        ('description', 500),
        ('comments', 1000),
        ('status', 1000),
    )



# Generated by Django 5.0.9 on 2024-10-07 11:49

import django.db.models.deletion
import taggit.managers
import timezone_field.fields
import utilities.json
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('circuits', '0044_circuit_groups'),
        ('dcim', '0191_module_bay_rebuild'),
        ('extras', '0121_customfield_related_object_filter'),
    ]

    operations = [
        migrations.CreateModel(
            name='PhoneDelivery',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False)),
                ('created', models.DateTimeField(auto_now_add=True, null=True)),
                ('last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('custom_field_data', models.JSONField(blank=True, default=dict, encoder=utilities.json.CustomFieldJSONEncoder)),
                ('delivery', models.CharField(blank=True, null=True)),
                ('channel_count', models.PositiveBigIntegerField(blank=True, null=True)),
                ('status', models.CharField(max_length=30)),
                ('ndi', models.PositiveBigIntegerField(blank=True, null=True)),
                ('dto', models.PositiveBigIntegerField(blank=True, null=True)),
                ('description', models.CharField(blank=True, max_length=200)),
                ('comments', models.TextField(blank=True)),
                ('provider', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.RESTRICT, to='circuits.provider')),
                ('site', models.ForeignKey(on_delete=django.db.models.deletion.RESTRICT, to='dcim.site')),
                ('tags', taggit.managers.TaggableManager(through='extras.TaggedItem', to='extras.Tag')),
            ],
            options={
                'verbose_name': 'Phone Delivery',
                'verbose_name_plural': 'Phone Deliveries',
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='PhoneDID',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False)),
                ('created', models.DateTimeField(auto_now_add=True, null=True)),
                ('last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('custom_field_data', models.JSONField(blank=True, default=dict, encoder=utilities.json.CustomFieldJSONEncoder)),
                ('start', models.PositiveBigIntegerField(blank=True)),
                ('end', models.PositiveBigIntegerField(blank=True)),
                ('delivery', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='sop_phone.phonedelivery')),
                ('site', models.ForeignKey(blank=True, on_delete=django.db.models.deletion.CASCADE, to='dcim.site')),
                ('tags', taggit.managers.TaggableManager(through='extras.TaggedItem', to='extras.Tag')),
            ],
            options={
                'verbose_name': 'DID Range',
                'verbose_name_plural': 'DIDs',
                'ordering': ('start',),
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='PhoneMaintainer',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False)),
                ('created', models.DateTimeField(auto_now_add=True, null=True)),
                ('last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('custom_field_data', models.JSONField(blank=True, default=dict, encoder=utilities.json.CustomFieldJSONEncoder)),
                ('description', models.CharField(blank=True, max_length=200)),
                ('comments', models.TextField(blank=True)),
                ('name', models.CharField(unique=True)),
                ('slug', models.SlugField(blank=True, max_length=100, unique=True)),
                ('status', models.CharField(default='Unknown', max_length=30)),
                ('physical_address', models.CharField(blank=True, max_length=200, null=True)),
                ('shipping_address', models.CharField(blank=True, max_length=200, null=True)),
                ('time_zone', timezone_field.fields.TimeZoneField(blank=True, null=True)),
                ('latitude', models.DecimalField(blank=True, decimal_places=6, max_digits=8, null=True)),
                ('longitude', models.DecimalField(blank=True, decimal_places=6, max_digits=9, null=True)),
                ('tags', taggit.managers.TaggableManager(through='extras.TaggedItem', to='extras.Tag')),
            ],
            options={
                'verbose_name': 'Phone Maintainer',
                'verbose_name_plural': 'Phone Maintainers',
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='PhoneInfo',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False)),
                ('created', models.DateTimeField(auto_now_add=True, null=True)),
                ('last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('custom_field_data', models.JSONField(blank=True, default=dict, encoder=utilities.json.CustomFieldJSONEncoder)),
                ('site', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, to='dcim.site')),
                ('tags', taggit.managers.TaggableManager(through='extras.TaggedItem', to='extras.Tag')),
                ('maintainer', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='sop_phone.phonemaintainer')),
            ],
            options={
                'verbose_name': 'Information',
                'verbose_name_plural': 'Informations',
                'abstract': False,
            },
        ),
        migrations.AddConstraint(
            model_name='phonedelivery',
            constraint=models.UniqueConstraint(fields=('ndi',), name='sop_phone_phonedelivery_unique_ndi', violation_error_message='NDI must be unique.'),
        ),
        migrations.AddConstraint(
            model_name='phonedelivery',
            constraint=models.UniqueConstraint(fields=('delivery', 'site'), name='sop_phone_phonedelivery_unique_delivery_method_site', violation_error_message='Delivery method must be unique in a site.'),
        ),
        migrations.AddConstraint(
            model_name='phonedid',
            constraint=models.UniqueConstraint(fields=('start',), name='sop_phone_phonedid_unique_start_number', violation_error_message='Start number must be unique.'),
        ),
        migrations.AddConstraint(
            model_name='phonedid',
            constraint=models.UniqueConstraint(fields=('end',), name='sop_phone_phonedid_unique_end_number', violation_error_message='End number must be unique.'),
        ),
        migrations.AddConstraint(
            model_name='phonedid',
            constraint=models.CheckConstraint(check=models.Q(('end__gte', models.F('start')), ('start__biabs__bilog__bifloor', models.F('end'))), name='sop_phone_phonedid_end_greater_than_start', violation_error_message='End number must be greater than or equal to start number.'),
        ),
        migrations.AddConstraint(
            model_name='phonemaintainer',
            constraint=models.UniqueConstraint(fields=('name',), name='sop_phone_phonemaintainer_unique_name', violation_error_message='Maintainer name must be unique.'),
        ),
        migrations.AddConstraint(
            model_name='phonemaintainer',
            constraint=models.UniqueConstraint(fields=('slug',), name='sop_phone_phonemaintainer_unique_slug', violation_error_message='Maintainer slug must be unique.'),
        ),
        migrations.AddConstraint(
            model_name='phoneinfo',
            constraint=models.UniqueConstraint(fields=('site',), name='sop_phone_phoneinfo_unique_site', violation_error_message='Site must be unique.'),
        ),
    ]



from django.contrib.auth.mixins import PermissionRequiredMixin
from django.utils.translation import gettext_lazy as _
from django.urls import reverse
from django.shortcuts import get_object_or_404

from netbox.views import generic
from dcim.models import Site

from ..forms.phone_did import *
from ..tables.phone_did import *
from ..filtersets import PhoneDIDFilterSet
from ..models import *
from ..utils import count_all_did, format_number


__all__ = (
    'PhoneDIDEditView',
    'PhoneDIDDeleteView',
    'PhoneDIDDetailView',
    'PhoneDIDBulkEditView',
    'PhoneDIDBulkDeleteView',
    'PhoneDIDListView',
    'PhoneDIDBulkImportView',
    'PhoneDIDAddSiteView'
)


class PhoneDIDListView(generic.ObjectListView):
    '''
    all DIDs list
    '''
    queryset = PhoneDID.objects.all()
    table = PhoneDIDTable
    filterset_form = PhoneDIDFilterForm
    filterset = PhoneDIDFilterSet


class PhoneDIDBulkEditView(generic.BulkEditView):
    '''
    for the "edit selected" view
    '''
    queryset = PhoneDID.objects.all()
    table = PhoneDIDTable
    form = PhoneDIDBulkEditForm
    filterset = PhoneDIDFilterSet


class PhoneDIDBulkDeleteView(generic.BulkDeleteView):
    '''
    for the "delete selected" view
    '''
    queryset = PhoneDID.objects.all()
    table = PhoneDIDTable
    filterset = PhoneDIDFilterSet


class PhoneDIDEditView(generic.ObjectEditView):
    '''
    edits a DID instance
    '''
    queryset = PhoneDID.objects.all()
    form = PhoneDIDForm
    

class PhoneDIDDeleteView(generic.ObjectDeleteView):
    '''
    deletes a DID instance
    '''
    queryset = PhoneDID.objects.all()


class PhoneDIDDetailView(generic.ObjectView, PermissionRequiredMixin):
    '''
    returns the DID detail page with context
    '''
    queryset = PhoneDID.objects.all()

    def get_extra_context(self, request, instance):
        context: dict = {}

        context['start'] = format_number(instance.start)
        context['end'] = format_number(instance.end)
        context['num_did'] = count_all_did(instance).__int__()
        try:
            context['maintainer'] = PhoneInfo.objects.filter(site=instance.delivery.site).first()
        except:
            pass
        return context


class PhoneDIDBulkImportView(generic.BulkImportView):
    queryset = PhoneDID.objects.all()
    model_form = PhoneDIDBulkImportForm

    def save_object(self, object_form, request):
        instance = object_form.save()
        
        if not instance.end or instance.end == 0:
            instance.end = instance.start
            instance.save()

        return instance

    def post(self, request):
        '''
        post request handler
        if additionnal changes is needed
        '''
        response = super().post(request)
        return response


class PhoneDIDAddSiteView(generic.ObjectEditView):
    '''
    adds a site automatically to the DID
    '''
    queryset = PhoneDID.objects.all()
    form = PhoneDIDForm

    def get_object(self, **kwargs):
        return self.queryset.model(site=get_object_or_404(Site, pk=kwargs['pk']))

    def alter_object(self, obj, request, args, kwargs):
        pk = kwargs.get('pk')
        site = get_object_or_404(Site, pk=pk)
        obj = self.queryset.model
        return obj(site=site)

    def get(self, request, *args, **kwargs):
        '''
        get request handler
        '''
        response = super().get(request, *args, **kwargs)
        return response



from dcim.models import Site
from netbox.views import generic
from tenancy.views import ObjectContactsView
from utilities.views import GetRelatedModelsMixin, register_model_view

from ..models import PhoneMaintainer, PhoneInfo, PhoneDID, PhoneDelivery
from ..filtersets import PhoneMaintainerFilterSet
from ..tables.phone_maintainer import *
from ..forms.phone_maintainer import *
from ..utils import count_all_did


__all__ = (
    'PhoneMaintainerView',
    'PhoneMaintainerEditView',
    'PhoneMaintainerDeleteView',
    'PhoneMaintainerBulkEditView',
    'PhoneMaintainerBulkDeleteView',
    'PhoneMaintainerBulkImportView',
    'PhoneMaintainerContactsView'
)


class PhoneMaintainerListView(generic.ObjectListView):
    queryset = PhoneMaintainer.objects.all()
    table = PhoneMaintainerTable
    filterset = PhoneMaintainerFilterSet
    filterset_form = PhoneMaintainerFilterForm


class PhoneMaintainerView(generic.ObjectView, GetRelatedModelsMixin):
    queryset = PhoneMaintainer.objects.all()

    def count_did(self, sites) -> tuple[int, int]:
        '''
        num_did = count of all numbers
        '''
        num_did: int = 0

        for instance in sites:
            temp = count_all_did(
                PhoneDID.objects.filter(delivery__site=instance.site),
                PhoneDelivery.objects.filter(site=instance.site)
            )
            num_did += temp.__int__()

        return num_did
    
    def get_format(self, values) -> str | None:
        qs = [str(item['site__id']) for item in values]
        if qs == []:
            return None
        return f'id=' + '&id='.join(qs)

    def get_extra_context(self, request, instance):
        '''
        additionnal context for the related models/objects
        as they are not directly related
        '''
        context: dict = {}

        sites = PhoneInfo.objects.filter(maintainer=instance)
        site_ids = (PhoneInfo.objects.filter(maintainer=instance).values('site__id'))

        context['num_did'] = self.count_did(sites)
        context['site_ids'] = site_ids
        context['related_models'] = self.get_related_models(
            request, 
            instance, 
            extra=(
                (Site.objects.filter(
                    pk__in=(PhoneInfo.objects.filter(maintainer=instance).values('site__id'))
                ), 'id'),
                (PhoneDID.objects.filter(
                    delivery__site__in=PhoneInfo.objects.filter(maintainer=instance).values('site_id')
                ), 'maintainer_id')
            )
        )
        context['site'] = Site
        context['restricted'] = self.get_format(site_ids)
        return context


class PhoneMaintainerEditView(generic.ObjectEditView):
    '''
    edits a maintainer instance
    '''
    queryset = PhoneMaintainer.objects.all()
    form = PhoneMaintainerForm


class PhoneMaintainerDeleteView(generic.ObjectDeleteView):
    '''
    deletes a maintainer instance
    '''
    queryset = PhoneMaintainer.objects.all()


class PhoneMaintainerBulkDeleteView(generic.BulkDeleteView):
    '''
    deletes multiple phone maintainers instances
    '''
    queryset = PhoneMaintainer.objects.all()
    table = PhoneMaintainerTable
    filterset = PhoneMaintainerFilterSet


class PhoneMaintainerBulkEditView(generic.BulkEditView):
    '''
    edits multiple phone maintainer instances
    '''
    queryset = PhoneMaintainer.objects.all()
    table = PhoneMaintainerTable
    form = PhoneMaintainerBulkEditForm
    filterset = PhoneMaintainerFilterSet


class PhoneMaintainerBulkImportView(generic.BulkImportView):
    queryset = PhoneMaintainer.objects.all()
    model_form = PhoneMaintainerBulkImportForm

    def save_object(self, object_form, request):
        instance = object_form.save()
        return instance

    def post(self, request):
        '''
        post request handler
        if additionnal changes is needed
        '''
        response = super().post(request)
        return response


@register_model_view(PhoneMaintainer, 'contacts')
class PhoneMaintainerContactsView(ObjectContactsView):
    queryset = PhoneMaintainer.objects.all()



from django import forms
from django.utils.translation import gettext_lazy as _

from netbox.forms import NetBoxModelBulkEditForm, NetBoxModelForm, NetBoxModelFilterSetForm
from utilities.forms.fields import DynamicModelChoiceField
from utilities.forms.rendering import FieldSet
from dcim.models import Site, Region, SiteGroup

from ..models import PhoneInfo, PhoneMaintainer


__all__ = (
    'PhoneInfoForm',
    'PhoneInfoFilterForm',
    'PhoneInfoBulkEditForm',
)


class PhoneInfoFilterForm(NetBoxModelFilterSetForm):
    model = PhoneInfo
    site_id = DynamicModelChoiceField(
        queryset=Site.objects.all(),
        required=False,
        label=_('Site')
    )
    region_id = DynamicModelChoiceField(
        queryset=Region.objects.all(),
        required=False,
        label=_('Region')
    )
    group_id = DynamicModelChoiceField(
        queryset=SiteGroup.objects.all(),
        required=False,
        label=_('Site group')
    )
    maintainer_id = forms.ModelChoiceField(
        queryset=PhoneMaintainer.objects.all(),
        required=False,
        label=_('Maintainer')
    )
    
    fieldsets = (
        FieldSet(
            'region_id', 'group_id', 'site_id',
            name=_('Location')
        ),
        FieldSet(
            'maintainer_id',
            name=_('Attributes')
        )
    )


class PhoneInfoForm(NetBoxModelForm):

    site = forms.ModelChoiceField(
        queryset=Site.objects.all(),
        required=False,
        label=_('Site')
    )
    maintainer = forms.ModelChoiceField(
        queryset=PhoneMaintainer.objects.all(),
        required=False,
        label=_('Maintainer')
    )

    class Meta:
        model = PhoneInfo
        fields = ('site', 'maintainer', )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
        if 'tags' in self.fields:
            del self.fields['tags']


class PhoneInfoBulkEditForm(NetBoxModelBulkEditForm):
    model = PhoneInfo
    maintainer = forms.ModelChoiceField(
        queryset=PhoneMaintainer.objects.all(),
        required=False,
        label=_('Maintainer')
    )

    class Meta:
        fields = ('maintainer', )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if 'add_tags' in self.fields:
            del self.fields['add_tags']
        if 'remove_tags' in self.fields:
            del self.fields['remove_tags']



import django_tables2 as tables
from django.utils.translation import gettext_lazy as _

from netbox.tables import NetBoxTable

from ..models import PhoneInfo


__all__ = (
    'PhoneInfoTable',
)


class PhoneInfoTable(NetBoxTable):
    '''
    table for all Phone Info
    '''
    site = tables.Column(
        verbose_name=_('Site'), linkify=True
    )
    maintainer = tables.Column(
        verbose_name=_('Maintainer'), linkify=True
    )

    class Meta(NetBoxTable.Meta):
        model = PhoneInfo
        fields = ('pk', 'id', 'site', 'maintainer', 'created', 'last_updated')
        default_columns = ('id', 'site', 'maintainer', 'created', 'last_updated')



from netbox.plugins import PluginConfig


class SopPhoneConfig(PluginConfig):
    name = "sop_phone"
    verbose_name = "SOP Phone"
    description = "Manage phone informations of each site."
    version='0.3.5'
    author = "Leorevoir"
    author_email = "leoquinzler@epitech.eu"
    base_url = "sop-phone"

config = SopPhoneConfig



import re
import phonenumbers
from phonenumbers import NumberParseException

from django.utils.translation import gettext_lazy as _
from django.core.exceptions import ValidationError


__all__ = (
    'PhoneValidator',
    'number_quicksearch',
    'PhoneMaintainerValidator',
)


class PhoneMaintainerValidator:

    @staticmethod
    def check_address(address, name:str) -> None:
        if not address:
            return
        if address.strip() == '':
            return
        if not re.match("^([^\n]* \r?\n)+[^\n]+$", address):
            raise ValidationError({
                f'{name}': _("Address must be multiline and each line must end with a space.")
            })


class PhoneValidator:

    @staticmethod
    def check_site(site) -> None:
        if site is None:
            raise ValidationError({
                'site': _("Site must be set.")
            })

    @staticmethod
    def check_delivery_site(delivery, site) -> None:
        if delivery and site and delivery.site != site:
            raise ValidationError({
                'delivery': _("Delivery must be set to the same site as the DID.")
            })

    @staticmethod
    def check_delivery_overlaps(delivery, start:int, end:int) -> None:
        if not delivery:
            return
        lstart = len(str(start))
        if len(str(delivery.dto)) == lstart:
            if start <= delivery.dto and end >= delivery.dto:
                raise ValidationError({
                    'start': _(f'This range {start} -> {end} overlaps its own delivery DTO.')
                })

    @staticmethod
    def check_number(where:str, number:int) -> None:
        if number is None or number <= 0 :
            raise ValidationError({
                f'{where}': _("Number must be set in E164 format.")
            })
        try:
            if not phonenumbers.parse(f'+{number}'):
                raise ValidationError({
                    f'{where}': _("Number must be a valid phone number written in E164 format.")
                })
        except NumberParseException:
            raise ValidationError({
                f'{where}': _('Number must be a valid phone number written in E164 format')
                })

    @staticmethod
    def check_start_end(start:int, end:int) -> None:
        if start is None or end is None:
            return
        if len(str(start))!=len(str(end)):
            raise ValidationError({
                'end': _("End number must be the same length as start number.")
            })
        if start > end:
            raise ValidationError({
                'end': _("End number must be greater than or equal to the start number.")
            })


def number_quicksearch(start: int, end: int, pattern: str) -> bool:
    '''
    number quicksearch with pattern in range of start>>end
    '''
    pattern_len = len(pattern)
    pattern_int = int(pattern)

    # calculate the largest multiple of 10^len(pattern) <= start
    base = start - (start % 10 ** pattern_len)

    while base <= end:
        # check if the base + pattern is in the range
        if base + pattern_int >= start and base + pattern_int <= end:
            return True

        # move to the next block
        base += 10 ** pattern_len 

        # if the pattern search is impossible because base is too big
        # break the loop and return false
        if base > end - pattern_int:
            break

    return False




from netbox.api.viewsets import NetBoxModelViewSet
from netbox.api.metadata import ContentTypeMetadata

from ..models import *
from ..filtersets import *
from .serializers import *


__all__ = (
    'PhoneDeliveryViewSet',
    'PhoneDIDViewSet',
    'PhoneInfoViewSet',
    'PhoneMaintainerViewSet',
)

class PhoneMaintainerViewSet(NetBoxModelViewSet):
    metadata_class = ContentTypeMetadata
    queryset = PhoneMaintainer.objects.all()
    serializer_class = PhoneMaintainerSerializer
    filterset_class = PhoneMaintainerFilterSet


class PhoneInfoViewSet(NetBoxModelViewSet):
    metadata_class = ContentTypeMetadata
    queryset = PhoneInfo.objects.all()
    serializer_class = PhoneInfoSerializer
    filterset_class = PhoneInfoFilterSet


class PhoneDIDViewSet(NetBoxModelViewSet):
    metadata_class = ContentTypeMetadata
    queryset = PhoneDID.objects.all()
    serializer_class = PhoneDIDSerializer
    filterset_class = PhoneDIDFilterSet


class PhoneDeliveryViewSet(NetBoxModelViewSet):
    metadata_class = ContentTypeMetadata
    queryset = PhoneDelivery.objects.all()
    serializer_class = PhoneDeliverySerializer
    filterset_class = PhoneDeliveryFilterSet



from django.utils.translation import gettext_lazy as _
from django.shortcuts import get_object_or_404

from netbox.views import generic
from dcim.models import Site

from ..utils import count_all_did
from ..forms.phone_info import *
from ..models import PhoneInfo, PhoneDID
from ..filtersets import PhoneInfoFilterSet
from ..tables.phone_info import PhoneInfoTable


__all__ = (
    'PhoneInfoEditView',
    'PhoneInfoDeleteView',
    'PhoneInfoDetailView',
    'PhoneInfoListView',
    'PhoneInfoBulkEditView',
    'PhoneInfoBulkDeleteView',
    'PhoneInfoAddView',
)


class PhoneInfoAddView(generic.ObjectEditView):
    queryset = PhoneInfo.objects.all()
    form = PhoneInfoForm

    def get_object(self, **kwargs):
        '''
        if pk in kwargs, this the request is from "phone" site tab view
        so pk is the site id
        '''
        if 'pk' in kwargs:
            return self.queryset.model(site=get_object_or_404(Site, pk=kwargs['pk']))
        return self.queryset.model()

    def alter_object(self, obj, request, args, kwargs):
        '''
        '''
        if 'pk' in kwargs:
            pk = kwargs.get('pk')
            site = get_object_or_404(Site, pk=pk)
            obj = self.queryset.model
            return obj(site=site)
        return obj

    def get(self, request, *args, **kwargs):
        '''
        get request handler
        '''
        response = super().get(request, *args, **kwargs)
        return response


class PhoneInfoListView(generic.ObjectListView):
    queryset = PhoneInfo.objects.all()
    table = PhoneInfoTable
    filterset = PhoneInfoFilterSet
    filterset_form = PhoneInfoFilterForm


class PhoneInfoDetailView(generic.ObjectView):
    queryset = PhoneInfo.objects.all()
    
    def get_extra_context(self, request, instance) -> dict:
        context:dict = {}
        
        dids = PhoneDID.objects.filter(site=get_object_or_404(Site, pk=instance.site.id))
        context['num_did'] = count_all_did(dids).__int__()
        return context


class PhoneInfoEditView(generic.ObjectEditView):
    queryset = PhoneInfo.objects.all()
    form = PhoneInfoForm

    def get_return_url(self, request, obj=None):
        try:
            return '/dcim/sites/' + str(obj.site.pk) + '/phone/'
        except:
            return '/dcim/sites/'


class PhoneInfoDeleteView(generic.ObjectDeleteView):
    queryset = PhoneInfo.objects.all()

    def get_return_url(self, request, obj=None) -> str:
        try:
            if obj is None:
                raise Exception
            return f'/dcim/sites/{obj.site.pk}/phone'
        except:
            return '/dcim/sites/'


class PhoneInfoBulkEditView(generic.BulkEditView):
    queryset = PhoneInfo.objects.all()
    table = PhoneInfoTable
    form = PhoneInfoBulkEditForm
    filterset = PhoneInfoFilterSet


class PhoneInfoBulkDeleteView(generic.BulkDeleteView):
    queryset = PhoneInfo.objects.all()
    table = PhoneInfoTable
    filterset = PhoneInfoFilterSet
